#include <pspsdk.h>
#include <pspkernel.h>

#include <string.h>
#include <stdio.h>


PSP_MODULE_INFO("DowndaterTest_Helper", 0x1000, 1, 1);

PSP_MAIN_THREAD_ATTR(0);

PSP_HEAP_SIZE_KB(0);

#define printf pspDebugScreenPrintf

#define N_MODULES	7

#define PSP_SIGNATURE	0x5053507E

char *savepaths[N_MODULES] =
{
	"ms0:/DOWNDATER/PRX/libpsar.prx",
	"ms0:/DOWNDATER/PRX/chkuppkg.prx",
	"ms0:/DOWNDATER/PRX/lflash_fatfmt.prx",
	"ms0:/DOWNDATER/PRX/testmode.prx",
	"ms0:/DOWNDATER/PRX/lepton_update.prx",
	"ms0:/DOWNDATER/PRX/ipl_update.prx",
	"ms0:/DOWNDATER/PRX/suspend_canceler.prx"
};

u32 offsets[N_MODULES] =
{
	0x2f5100,
	0x2f71c0,
	0x2f8dc0,
	0x2fc180,
	0x301200,
	0x303080,
	0x33b500
};

void create_dirs()
{
	sceIoMkdir("ms0:/DOWNDATER", 0777);
	sceIoMkdir("ms0:/DOWNDATER/PRX", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/data", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/data/cert", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/dic", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/font", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/kd", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/kd/resource", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/vsh", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/vsh/etc", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/vsh/module", 0777);
	sceIoMkdir("ms0:/DOWNDATER/DUMP/vsh/resource", 0777);
}

SceUID outputfilelist=-1, inputfilelist=-1;

void ErrorExit(char *error)
{
	printf("%s\n", error);

	sceKernelDelayThread(4 * 1000 * 1000);
	sceKernelExitGame();	
}

void writeOutputListEntry(char *str)
{
	char entry[1024];

	strcpy(entry, str);

	entry[5] = '8';
	
	sceIoWrite(outputfilelist, entry, strlen(entry)+1);
}

void writeInputListEntry(char *str)
{
	sceIoWrite(inputfilelist, str, strlen(str)+1);
}

char buffer[32*1024];

void copyfile(char *input, char *output)
{
	SceUID infd = sceIoOpen(input, PSP_O_RDONLY, 0777);
	SceUID outfd = sceIoOpen(output, PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);

	int bytesread;

	printf("copying %s to %s\n", input, output);

	if (infd < 0 || outfd < 0)
	{
		printf("Error copying file");
		ErrorExit("");
	}

	while ((bytesread = sceIoRead(infd, buffer, 32*1024)) > 0)
	{
		sceIoWrite(outfd, buffer, bytesread);
	}

	sceIoClose(infd);
	sceIoClose(outfd);
}

char inputpath[1024], outputpath[1024];

void dump_dir(char *path)
{
	int dfd;

	dfd = sceIoDopen(path);

	if (dfd < 0)
	{
		printf("Error opeining dir %s\n", path);
		ErrorExit("");
		return;
	}

	SceIoDirent dir;

	while(sceIoDread(dfd, &dir) > 0)
	{
		if(dir.d_stat.st_attr & FIO_SO_IFDIR)
		{
			if(dir.d_name[0] != '.')
			{
				char nextdir[1024];

				sprintf(nextdir, "%s%s/", path, dir.d_name);

				dump_dir(nextdir);
			}
		}
		else
		{
			sprintf(inputpath, "%s%s", path, dir.d_name);
			sprintf(outputpath, "ms0:/DOWNDATER/DUMP/%s", inputpath+8);
			writeOutputListEntry(inputpath); // Input now, output after
			writeInputListEntry(outputpath); // Output now, input after
			copyfile(inputpath, outputpath);
		}
	}
}

void dump_flash()
{

	outputfilelist = sceIoOpen("ms0:/DOWNDATER/outputfl.bin", PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);
	inputfilelist = sceIoOpen("ms0:/DOWNDATER/inputfl.bin", PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);

	if (inputfilelist < 0 || outputfilelist < 0)
		ErrorExit("Error creating list files.");

	dump_dir("flash0:/");

	char null = 0;

	sceIoWrite(inputfilelist, &null, 1);
	sceIoClose(inputfilelist);

	sceIoWrite(outputfilelist, &null, 1);
	sceIoClose(outputfilelist);
}

void extract_prx()
{
	SceKernelModuleInfo modinfo;
	u32 base;
	int i;

	sceIoChdir("ms0:/PSP/GAME/UPDATE");
	
	// Load the update (without starting it) to extract the special prx's
	SceUID mod = sceKernelLoadModule("EBOOT.PBP", 0, NULL);

	if (mod < 0)
		ErrorExit("Error loading update module.\n");

	if (sceKernelQueryModuleInfo(mod, &modinfo) < 0)
		ErrorExit("Cannot query module info.\n");

	base = modinfo.text_addr;

	for (i = 0; i < N_MODULES; i++)
	{
		if (_lw(base + offsets[i]) != PSP_SIGNATURE)
		{
			ErrorExit("Cannot find module\n");
		}

		// Thanks http://ps2dev.ps2-scene.org/pspformat.txt for the psp header info ;)
		int size = _lw(base + offsets[i] + 0x2C);

		printf("Writing module %s size=%d\n", savepaths[i], size);

		SceUID fd = sceIoOpen(savepaths[i], PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);

		if (fd < 0)
			ErrorExit("Cannot save module.\n");
		
		printf("Writing module...\n\n");

		sceIoWrite(fd, (void *)(base + offsets[i]), size);

		sceIoClose(fd);
	}
}




int main()
{
	pspDebugScreenInit();

	printf("Creating dirs.\n");
	create_dirs();

	printf("Extracting special PRX's from the update...\n");
	extract_prx();

	printf("Dumping flash0...\n");
	dump_flash();



	ErrorExit("Finished. Exiting in four seconds\n");

	return 0;
}

